# Generated by Django 3.0.8 on 2020-08-05 09:33

import django.db.models.deletion
import django.utils.timezone

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_localize", "0003_delete_translation_sources"),
    ]

    operations = [
        migrations.AddField(
            model_name="translationsource",
            name="last_updated_at",
            field=models.DateTimeField(default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="translationsource",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterUniqueTogether(
            name="translationsource",
            unique_together={("object", "locale")},
        ),
    ]
