{% extends 'wagtailadmin/shared/headers/slim_header.html' %}
{% load wagtailadmin_tags i18n %}

{% block header_content %}
    <h1 class="sr-only">
        {% blocktrans trimmed with title=instance object_type=model_opts.verbose_name %}Editing {{ object_type }} {{ title }}{% endblocktrans %}
    </h1>
{% endblock %}

{% block actions %}
    {% with nav_icon_classes='w-w-4 w-h-4' nav_icon_button_classes='w-h-slim-header w-bg-transparent w-box-border w-py-3 w-px-3 w-flex w-justify-center w-items-center w-outline-offset-inside w-text-grey-400 w-transition hover:w-transform hover:w-scale-110 hover:w-text-primary focus:w-text-primary' %}
        {% include "wagtailadmin/shared/side_panel_toggles.html" %}
    {% endwith %}
{% endblock %}
